#! /usr/bin/python3
#
#################################################################
## Copyright (c) 2016 Norbert S. <junky-zs@gmx.de>
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#
#################################################################
# Ver:0.0.1  / Datum 03.03.2016
#################################################################
#
# This tool takes binary logfiles (heater-type: heatronic) and
# formating it as good as possible to a csv-file written to the
# current folder with name: '<filename>.message.csv'.
# 
# There is implemented a detection of ht_transceiver messages.
# If available, this tool uses that header-informations regarding
#  length of the message-included payload.
# If this information is NOT available the detection is made with
#  CRC-check and searching for valid messages.
#  Keep in mind this detection-type can have wrong detections,
#  cause limited CRC(8)-check.
#  Also there are polling messages included without CRC-check.
#  So the written crc-file has to be carefully checked for valid 
#  messages.
# 
#################################################################

import sys

class cht_bin2csv(object):
    def __init__(self, filename):
        
        self.__crc_table = [ \
            0x00, 0x02 ,0x04, 0x06, 0x08, 0x0A, 0x0C, 0x0E, 0x10, 0x12, 0x14, 0x16, 0x18,
            0x1A, 0x1C, 0x1E, 0x20, 0x22, 0x24, 0x26, 0x28, 0x2A, 0x2C, 0x2E, 0x30, 0x32,
            0x34, 0x36, 0x38, 0x3A, 0x3C, 0x3E, 0x40, 0x42, 0x44, 0x46, 0x48, 0x4A, 0x4C,
            0x4E, 0x50, 0x52, 0x54, 0x56, 0x58, 0x5A, 0x5C, 0x5E, 0x60, 0x62, 0x64, 0x66,
            0x68, 0x6A, 0x6C, 0x6E, 0x70, 0x72, 0x74, 0x76, 0x78, 0x7A, 0x7C, 0x7E, 0x80,
            0x82, 0x84, 0x86, 0x88, 0x8A, 0x8C, 0x8E, 0x90, 0x92, 0x94, 0x96, 0x98, 0x9A,
            0x9C, 0x9E, 0xA0, 0xA2, 0xA4, 0xA6, 0xA8, 0xAA, 0xAC, 0xAE, 0xB0, 0xB2, 0xB4,
            0xB6, 0xB8, 0xBA, 0xBC, 0xBE, 0xC0, 0xC2, 0xC4, 0xC6, 0xC8, 0xCA, 0xCC, 0xCE,
            0xD0, 0xD2, 0xD4, 0xD6, 0xD8, 0xDA, 0xDC, 0xDE, 0xE0, 0xE2, 0xE4, 0xE6, 0xE8,
            0xEA, 0xEC, 0xEE, 0xF0, 0xF2, 0xF4, 0xF6, 0xF8, 0xFA, 0xFC, 0xFE, 0x19, 0x1B,
            0x1D, 0x1F, 0x11, 0x13, 0x15, 0x17, 0x09, 0x0B, 0x0D, 0x0F, 0x01, 0x03, 0x05,
            0x07, 0x39, 0x3B, 0x3D, 0x3F, 0x31, 0x33, 0x35, 0x37, 0x29, 0x2B, 0x2D, 0x2F,
            0x21, 0x23, 0x25, 0x27, 0x59, 0x5B, 0x5D, 0x5F, 0x51, 0x53, 0x55, 0x57, 0x49,
            0x4B, 0x4D, 0x4F, 0x41, 0x43, 0x45, 0x47, 0x79, 0x7B, 0x7D, 0x7F, 0x71, 0x73,
            0x75, 0x77, 0x69, 0x6B, 0x6D, 0x6F, 0x61, 0x63, 0x65, 0x67, 0x99, 0x9B, 0x9D,
            0x9F, 0x91, 0x93, 0x95, 0x97, 0x89, 0x8B, 0x8D, 0x8F, 0x81, 0x83, 0x85, 0x87,
            0xB9, 0xBB, 0xBD, 0xBF, 0xB1, 0xB3, 0xB5, 0xB7, 0xA9, 0xAB, 0xAD, 0xAF, 0xA1,
            0xA3, 0xA5, 0xA7, 0xD9, 0xDB, 0xDD, 0xDF, 0xD1, 0xD3, 0xD5, 0xD7, 0xC9, 0xCB,
            0xCD, 0xCF, 0xC1, 0xC3, 0xC5, 0xC7, 0xF9, 0xFB, 0xFD, 0xFF, 0xF1, 0xF3, 0xF5,
            0xF7, 0xE9, 0xEB, 0xED, 0xEF, 0xE1, 0xE3, 0xE5, 0xE7]

        self._rawdata = []
        self._max_messagesize = 39
        
        self._filehandle=None
        self._filehandleout=None
        self._inputfile=filename
        self._outputfile=filename+".message.csv"
        self._ht_transceiver_header_found = False

        self._maxcalls_perline = 50
        self._maxdots_perline  = 80
        self._current_call = 0
        self._current_dotcounts=0

        
        print("Filename:{0}".format(self._inputfile))
        try:
            self._filehandle=open(self._inputfile,"rb")
            self._filehandleout=open(self._outputfile,"w+")
            self._filehandleout.write("Date;Time;Size;B0;B1;B2;B3;B4;B5;B6;B7;B8;B9;10;11;12;13;14;15;16;17;18;19;20;21;22;23;24;25;26;27;28;29;30;31;32;33\n")
            self._filehandleout.flush()
        except:
            print('cht_bin2csv();Error; could not open file:{0}'.format(self._inputfile))
            quit()

    def __del__(self):
        if self._filehandle != None:
            self._filehandle.close()
            self._filehandleout.close()

    def _crc_testen(self, buffer, bufferlength, bufferstartindex):
        crc = 0
        crcvalid = False
        if len(buffer) > 2:
            try:
                for i in range(0, bufferlength):
                    crc = self.__crc_table[crc] 
                    crc ^= buffer[bufferstartindex+i]
                else:
                    if crc == buffer[bufferstartindex+bufferlength]:
                        crcvalid = True
                    else:
                        crcvalid = False
                return crcvalid
            except (IndexError) as e:
                print("cht_bin2csv();_crc_testen();Error;",e.args[0])
                return False
        else:
            return False
        
    def _read(self):
        try:
            value=self._filehandle.read(1)
        except:
            print("Error reading values from file")
            value=None
        return value
    
    def _read_rawdata(self):
        rtnvalue = 0
        if len(self._rawdata) < self._max_messagesize:
            #read data to rawbuffer
            readcounter = self._max_messagesize - len(self._rawdata)
            while readcounter > 0:
                value=self._read()
                if (value == None or len(value) == 0):
                    rtnvalue = -1
                    break
                else:
                    self._rawdata.append(ord(value))
                readcounter -= 1
        if not rtnvalue == -1 :
            rtnvalue = len(self._rawdata)
        return rtnvalue

    def _DateTimeMessage(self, buffer, bufferreadindex):
        info_datum=0
        info_zeit =0
        if (( buffer[bufferreadindex]==0x90 or self._rawdata[bufferreadindex]==0x10) and
              buffer[bufferreadindex + 1] == 0 and
              buffer[bufferreadindex + 2] == 6 and
              buffer[bufferreadindex + 3] ==0
            ):
            iyear      = int(buffer[bufferreadindex+4]+2000)
            imonth     = int(buffer[bufferreadindex+5])
            ihour      = int(buffer[bufferreadindex+6])
            iday       = int(buffer[bufferreadindex+7])
            iminute    = int(buffer[bufferreadindex+8])
            isecond    = int(buffer[bufferreadindex+9])
            idayofweek = int(buffer[bufferreadindex+10])
            
            info_datum="""{day:02}.{month:02}.{year:4}""".format(day=iday,
                                                                   month=imonth,
                                                                   year=iyear)
            info_zeit ="""{hour:02}:{minute:02}:{second:02}""".format(hour=ihour,
                                                                        minute=iminute,
                                                                        second=isecond)

            return "{0};{1};".format(info_datum,info_zeit)
        else:
            return ""

    def _ValidStartByte(self, startbyte):
        rtn = False
        if startbyte > 0x80:
            rtn = True
        return rtn

    ## check: is ht_transceiver Msg-header available ?
    def _Check_TransceiverMsgHeader(self, buffer):
        rtn_flag=False
        try:
            if (buffer[0] == 0x23): # start-tag '#'
                # looking for 'H' and 'R'
                if (buffer[1]==0x48 and buffer[2]==0x52):
                    rtn_flag=True
                else:
                    rtn_flag=False
        except:
            print("could not check MsgHeader")
        return rtn_flag

    def _print_inprogress_dots(self):
        self._current_call += 1
        if not (self._current_call % self._maxcalls_perline):
            print (".",end="")
            self._current_dotcounts += 1

        if not (self._current_dotcounts % self._maxdots_perline):
            self._current_dotcounts += 1
            print("")

    def run(self):
        value=0
        loop=True
        DateTimeString="0;0;"
        payload_size = 0
        message_size = 0
        check_msgheader_firsttime = True
        
        while loop:
            try:
                if not self._read_rawdata() > 0:
                    loop=False
                    quit
                try:
                    # search for ht_transceiver message-header
                    if check_msgheader_firsttime:
                        # search header from byte 0 to max-buffersize - header - minimum payload-bytes
                        for check_index in range(0, len(self._rawdata) - 8):
                            self._ht_transceiver_header_found = self._Check_TransceiverMsgHeader(self._rawdata[check_index:len(self._rawdata)])
                            if (self._ht_transceiver_header_found):
                                break
                        if (self._ht_transceiver_header_found):
                            self._max_messagesize = 39
                            print ("------> ht_transceiver-header found <------")
                        else:
                            self._max_messagesize = 33
                            print ("------> common CRC-detection procedure <------")
                        check_msgheader_firsttime = False
                except:
                    print ("cht_bin2csv.run();Error;messageheader search failed")
                    raise
                        
                try:
                    if (self._ht_transceiver_header_found == False):
                        while len(self._rawdata) > 0:
                            # search for message source start-tag (> 0x80)
                            if not self._ValidStartByte(self._rawdata[0]):
                                del self._rawdata[0]
                            else:
                                break
                except:
                    print ("cht_bin2csv.run();Error;start-tag search failed")
                    raise

                payload_size = 0
                message_size = 0
                if self._ht_transceiver_header_found == True:
                    #################################################
                    ### Handling for: ht_transceiver-header found ###
                    #################################################
                    # get payload-size from header and rest of msg (including terminating CRC) from file
                    #  msg-struct: #HR(11)h<size><payload-bytes><CRC-byte>
                    if self._Check_TransceiverMsgHeader(self._rawdata):
                        payload_size = self._rawdata[4]
                        # msg-size := payload-size + header-size + CRC-byte
                        message_size = payload_size + 6
                        # check current buffer-size, load the rest if less then msg-size 
                        if len(self._rawdata) < message_size:
                            readcounter = message_size - len(self._rawdata)
                            while readcounter > 0:
                                value=self._read()
                                if (value == None or len(value) == 0):
                                    break
                                else:
                                    self._rawdata.append(ord(value))
                                readcounter -= 1
                        # extract payload and write it to csv-file, suppress polling-messages (without CRC)
                        if (payload_size < 3):
                            # delete this from buffer
                            del self._rawdata[0:message_size]
                        else:
                            crc_check=self._crc_testen(self._rawdata, payload_size - 2, 5)
                            #check valid crc for that current raw-buffer content and the terminating Break-Sign := 0
                            if (crc_check and self._rawdata[4 + payload_size] == 0):
                                temptime = self._DateTimeMessage(self._rawdata, 5)
                                if len(temptime) > 0:
                                    DateTimeString = temptime
                                    
                                # write formated msg to csv-file
                                temptext=DateTimeString+"{0};".format(payload_size)
                                readindex = 5
                                while readindex < payload_size + 5:
                                    temptext += format(self._rawdata[readindex],"02x")+";"
                                    readindex += 1
                                temptext += "\n"
                                self._filehandleout.write(temptext)
                                self._filehandleout.flush()
                                # delete that rawdata from buffer
                                del self._rawdata[0:message_size]

                                self._print_inprogress_dots()
                            else:
                                if len(self._rawdata) > 0:
                                    del self._rawdata[0]
                    else:
                        if len(self._rawdata) > 0:
                            del self._rawdata[0]
                else:
                    ####################################################
                    ### Handling for: common CRC-detection procedure ###
                    ####################################################
                    try:
                        while len(self._rawdata) > 0:
                            # search for message source start-tag (> 0x80)
                            if not self._ValidStartByte(self._rawdata[0]):
                                del self._rawdata[0]
                            else:
                                break
                            
                        if len(self._rawdata) > 7:
                            if len(self._rawdata) < self._max_messagesize:
                                for crc_checklength in range(5, (len(self._rawdata) - 2)):
                                    crc_check=self._crc_testen(self._rawdata, crc_checklength, 0)
                                    #check valid crc for that current raw-buffer content and the terminating Break-Sign := 0
                                    if crc_check and self._rawdata[crc_checklength + 1] == 0:
                                        temptime = self._DateTimeMessage(self._rawdata, 0)
                                        if len(temptime) > 0:
                                            DateTimeString = temptime
                                            
                                        # currentlength plus 2 Byte (CRC- and Break-Byte)
                                        temptext=DateTimeString+"{0};".format(crc_checklength + 2)
                                        # write formated msg to csv-file
                                        readindex = 0
                                        while readindex < crc_checklength+2:
                                            temptext += format(self._rawdata[readindex],"02x")+";"
                                            readindex += 1
                                        temptext += "\n"
                                        self._filehandleout.write(temptext)
                                        self._filehandleout.flush()
                                        # delete that rawdata from buffer
                                        del self._rawdata[0:crc_checklength + 2]
                                        break
                                    self._print_inprogress_dots()
                            else:
                                # delete first byte in raw-buffer
                                del self._rawdata[0]
                        else:
                            # read more raw-data from file
                            pass
                    except:
                        print ("cht_bin2csv.run();Error;common search failed")
                        raise
            except (LookupError, IndexError, KeyError, TypeError, IOError, UnboundLocalError) as e:
                print('cht_bin2csv.run();Error;<{0}>'.format(e.args[0]))
                loop = False
        print("")
                
#--- class cht_bin2csv end ---#
################################################
            
filename = ""
if len(sys.argv) <= 1:
    print("Logfile Name erforderlich (Junkers-HT Format)")
else:
    filename = sys.argv[1]
    ht3_msgsearch=cht_bin2csv(filename)
    ht3_msgsearch.run()
